package model;

/*
 * Requirement:
 * You are required to implement all methods recursively.
 */
public class RecursiveMethods {
	/* https://codingbat.com/prob/p195413
	 * 
	 * Given a string and a non-empty substring sub, compute recursively the largest substring which starts and ends with sub and return its length.
	 * 
	 * strDist("catcowcat", "cat") → 9
	 * strDist("catcowcat", "cow") → 3
	 * strDist("cccatcowcatxx", "cat") → 9
	 * 
	 * Hint: 
	 * 	+ Keep track of the start indices of the first and last occurrences of the input `sub` string.
	 * 	+ Inspect the input string `str` from left to right. 
	 * 	+ When reaching the point where the remaining characters cannot contain `sub`, 
	 * 		calculate the max length according to where the first and last occurrences are.	
	 */
	public int strDist(String str, String sub) {
		/*
		 * Start scanning the input string `str` from index 0.
		 * Initially, NO occurrences of `sub` has been found in `str`:
		 * 	index of the start of the first occurrence
		 * 	and 
		 * 	index of the start of the last occurrence
		 * 	are both set to -1 to denote that NO occurrence has been found.
		 */
		return strDistHelper(str, sub, 0, -1, -1);
	}
	
	/*
	 * Recursive helper method:
	 * 	+ `str`: input string (call by value)
	 * 	+ `sub`: input substring (call by value)
	 * 	+ `index`: index of the current character to inspect (we scan from left to right)
	 * 	+ `startOfFirstIndex`: index of the starting character of the first occurrence of `sub`
	 * 	+ `startOfLastIndex`: index of the starting character of the first occurrence of `sub`
	 */
	private int strDistHelper(String str, String sub, int index, int startOfFirstOcc, int startOfLastOcc) {
		int subLength = sub.length();
		if(index > str.length() - subLength) { /* base case: the remaining characters cannot contain `sub` */
			if(startOfFirstOcc == -1 && startOfLastOcc == -1) {
				/* No occurrence of `sub` is found */
				return 0;
			}
			else {
				/* Calculate the max length according to where the first and last occurrences are. */
				return (startOfLastOcc + subLength) - startOfFirstOcc;
			}
		}
		else {
			/* A new occurrence of `sub` is found.
			 * Update the first and last occurrences accordingly. 
			 */
			if(str.substring(index, index + subLength).equals(sub)) {
				startOfLastOcc = index;
				if(startOfFirstOcc == -1) {
					startOfFirstOcc = index;
				}
			}
			/*
			 * Recur on a strictly smaller problem by considering the next character.
			 * Touch insight: `index + 1` here because the characters in the range of index, index + 1, ..., index + subLength - 1
			 * 	may contain the start of another occurrence. See testStrDist_6 (where one occurrence of `hih` is found at index 9, and another occurrence is already starting at index 11)
			 */
			return strDistHelper(str, sub, index + 1, startOfFirstOcc, startOfLastOcc);
		}
	}
}
